package cn.iocoder.yudao.module.member.mq.consumer.score.core;

import cn.hutool.core.collection.CollectionUtil;
import cn.hutool.core.util.ArrayUtil;
import cn.hutool.json.JSONUtil;
import cn.iocoder.boot.module.order.api.OrderApi;
import cn.iocoder.boot.module.order.api.dto.OrderItemRespDTO;
import cn.iocoder.boot.module.order.api.dto.OrderObjectiveApiDO;
import cn.iocoder.boot.module.order.api.dto.OrderRespDTO;
import cn.iocoder.boot.module.order.enums.OrderStatusApiEnum;
import cn.iocoder.yudao.module.customer.dal.dataobject.customerContacts.CustomerContactsDO;
import cn.iocoder.yudao.module.customer.service.customerContacts.CustomerContactsService;
import cn.iocoder.yudao.module.member.api.score.MemberUserScoreApi;
import cn.iocoder.yudao.module.member.api.score.dto.MemberUserScoreOperateReqDTO;
import cn.iocoder.yudao.module.member.dal.dataobject.scoreRule.ScoreRuleDO;
import cn.iocoder.yudao.module.member.enums.*;
import cn.iocoder.yudao.module.member.mq.message.ScoreMessage;
import cn.iocoder.yudao.module.member.service.scoreLog.MemberUserScoreLogService;
import cn.iocoder.yudao.module.member.service.scoreRule.ScoreRuleService;
import cn.iocoder.yudao.module.member.service.user.MemberUserService;
import cn.iocoder.yudao.module.member.vo.scoreRule.extra.ScoreRuleOrderVExtraVO;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.redisson.RedissonMultiLock;
import org.redisson.api.RLock;
import org.redisson.api.RedissonClient;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.math.BigDecimal;
import java.math.RoundingMode;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.TimeUnit;

/**
 * @author zhangfeng
 */
@Service
@Slf4j
public class OrderVStrategy extends AbstractScoreRuleStrategy {
    @Resource
    private OrderApi orderApi;
    @Resource
    private CustomerContactsService customerContactsService;
    @Resource
    private RedissonClient redissonClient;

    public OrderVStrategy(ScoreRuleService scoreRuleService, MemberUserScoreLogService memberUserScoreLogService, MemberUserService memberUserService, MemberUserScoreApi memberUserScoreApi) {
        super(scoreRuleService, memberUserScoreLogService, memberUserService, memberUserScoreApi);
    }

    @Override
    public void addScore(ScoreMessage message) {
        try {
            Thread.sleep(3000);
        } catch (InterruptedException e) {
            log.error("Order in shipping listener: Thread sleep error", e);
        }
        log.info("Received order V message,orderID:{}", message.getOrderId());
        Long orderId = message.getOrderId();
        // 获取订单信息
        OrderRespDTO order;
        try {
            order = orderApi.getOrder(orderId, message.getOrderNo());
        } catch (Exception e) {
            log.error("Order in shipping listener: Get order error,orderId:{}", orderId == null ? message.getOrderNo() : orderId, e);
            return;
        }
        if (Objects.isNull(order)) {
            log.error("Order in shipping listener: The order does not exist,orderId:{}", orderId == null ? message.getOrderNo() : orderId);
            return;
        }
        orderId = order.getOrderId();
        // 订单状态应为起运
        if (!Objects.equals(order.getStatus(), OrderStatusApiEnum.IN_SHIPPING.getValue())) {
            log.info("Order in shipping listener: The order status is not in shipping,orderID:{}", orderId);
            return;
        }
        // 获取订单规则
        Integer transportId = order.getTransportId();
        ScoreRuleDO scoreRuleDO = scoreRuleService.getEnableScoreRuleByType(ScoreRuleTypeEnum.ORDER_V, TransportTypeEnum.parseByValue(transportId));
        if (Objects.isNull(scoreRuleDO)) {
            log.info("Order in shipping listener: No score rule match,orderID:{}", orderId);
            return;
        }
        // 规则是否包含目的地提货点，0代表所有提货点
        OrderObjectiveApiDO orderObjectiveDO = order.getOrderObjectiveDO();
        ScoreRuleOrderVExtraVO extraInfo = JSONUtil.toBean(scoreRuleDO.getExtra(), ScoreRuleOrderVExtraVO.class);
        String[] warehouseIds = extraInfo.getReceiveAddr().split(",");
        if (!Objects.equals(warehouseIds[0], "0") && !ArrayUtil.contains(warehouseIds, orderObjectiveDO.getObjectiveWarehouseId().toString())) {
            log.info("Order in shipping listener: The rule does not contain the order destination warehouse,orderID:{},warehouseId:{}", orderId, orderObjectiveDO.getObjectiveWarehouseId().toString());
            return;
        }
        String lockKey1;
        if (order.getUserId() > 0) {
            lockKey1 = "score:rule:order:lock:" + order.getUserId();
        } else {
            lockKey1 = "score:rule:order:lock:" + order.getOrderConsignorDO().getCustomerId();
        }
        String lockKey2 = "score:rule:order:lock:" + order.getOrderConsigneeDO().getCustomerId();
        RLock lock1 = redissonClient.getLock(lockKey1);
        RLock lock2 = redissonClient.getLock(lockKey2);
        RedissonMultiLock multiLock = new RedissonMultiLock(lock1, lock2);
        try {
            boolean tryLock = multiLock.tryLock(2, 10, TimeUnit.SECONDS);
            if (!tryLock) {
                log.error("Order in shipping listener: get lock error, orderId:{}", orderId);
            }
            // 获取需要添加积分的用户 发货人
            List<Long> userIds = getAddScoreUser(order, extraInfo, scoreRuleDO);
            if (CollectionUtil.isEmpty(userIds)) {
                log.info("Order in shipping listener: consignor and consignee are not bound to the member, orderId:{}", orderId);
                return;
            }
            // 计算积分
            Integer scoreCount = getScoreCount(transportId, order.getChannelId(), extraInfo.getChannel(), extraInfo.getOrderVRule(), orderId);
            if (scoreCount == 0) {
                log.info("Order in shipping listener: The score count is less than or equal to 0,orderId:{}", orderId);
                return;
            }
            // 增加积分
            for (Long userId : userIds) {
                try {
                    HashMap<String, Object> map = new HashMap<>();
                    map.put("orderId", orderId);
                    memberUserScoreApi.operateScore(MemberUserScoreOperateReqDTO.builder()
                            .memberId(userId)
                            .scoreCount(scoreCount)
                            .sourceType(ScoreSourceTypeEnum.ORDER_V)
                            .ruleId(scoreRuleDO.getId())
                            .expireDays(scoreRuleDO.getScorePeriod())
                            .releationId(String.valueOf(orderId))
                            .uniqueId(orderId + "_" + userId)
                            .extParam(map)
                            .build()
                    );
                    log.info("Order V rule add score success,ruleID:{},orderID:{},userID:{}", scoreRuleDO.getId(), message.getOrderId(), userId);
                } catch (Exception e) {
                    log.error("Order in shipping listener: operateScore error", e);
                }
            }
        } catch (InterruptedException e) {
            log.error("Order in shipping listener: lock error orderId:{}", orderId, e);
        } finally {
            multiLock.unlock();
        }
    }

    /**
     * 根据规则获得要加多少分
     *
     * @param transportId
     * @param channelId
     * @param allowChannels
     * @param orderVRules
     * @param orderId
     * @return
     */
    private Integer getScoreCount(Integer transportId, Long channelId, String allowChannels, List<ScoreRuleOrderVExtraVO.OrderVRule> orderVRules, Long orderId) {
        Integer scoreCount = 0;
        List<OrderItemRespDTO> orderItems = orderApi.getOrderItemsByOrderId(orderId);
        if (orderItems == null) {
            log.info("Order in shipping listener: The order item is null,orderId:{}", orderId);
            return 0;
        }
        BigDecimal totalWeight = BigDecimal.ZERO;
        BigDecimal totalVolume = BigDecimal.ZERO;
        for (OrderItemRespDTO orderItem : orderItems) {
            totalWeight = totalWeight.add(orderItem.getChargeWeight());
            totalVolume = totalVolume.add(orderItem.getChargeVolume());
        }
        if (transportId == TransportTypeEnum.OCEAN_LCL.getValue()) {
            // 海运算体积
            scoreCount = computeScoreCount(totalVolume, orderVRules);
        } else if (transportId == TransportTypeEnum.SPECIAL_LINE_AIR_FREIGHT.getValue()) {
            // 空运校验渠道
            // 渠道id
            String[] channels = allowChannels.split(",");
            if (!ArrayUtil.contains(channels, channelId.toString())) {
                return 0;
            }
            // 空运算重量
            scoreCount = computeScoreCount(totalWeight, orderVRules);
        }
        return scoreCount;
    }

    /**
     * 获取需要加分的用户
     *
     * @param order
     * @param extraInfo
     * @param scoreRuleDO
     * @return
     */
    private List<Long> getAddScoreUser(OrderRespDTO order, ScoreRuleOrderVExtraVO extraInfo, ScoreRuleDO scoreRuleDO) {
        ArrayList<Long> users = new ArrayList<>();
        // 发货人和收货人都要判断给分
        Long userId = order.getUserId();
        Long consigneeId = order.getOrderConsigneeDO().getCustomerId();
        Long consigneeMemberId;
        boolean isBackendOrder = userId <= 0;
        // 校验订单入口
        if (!isOrderEntryValid(extraInfo.getOrderEntry(), isBackendOrder)) {
            return null;
        }
        // 如果是后台下单，重新获取发货人用户ID
        if (isBackendOrder) {
            userId = getCustomerMemberId(order.getOrderConsignorDO().getCustomerId());
            if (userId == null) {
                userId = 0L;
            }
        }
        // 校验首单
        if (!isFirstOrderValid(userId, extraInfo)) {
            return null;
        }
        // 校验积分上限，添加加分用户列表
        if (isBackendOrder && isUserScoreMaxed(userId, scoreRuleDO)) {
            log.info("Order in shipping listener: The user has reached the maximum score, userId:{}, scoreRuleId:{}", userId, scoreRuleDO.getId());
        } else {
            addValidUser(users, userId);
        }
        consigneeMemberId = getCustomerMemberId(consigneeId);
        if (consigneeMemberId != null && consigneeMemberId > 0) {
            if (consigneeMemberId.equals(userId)) {
                log.info("Order in shipping listener: The consignee is the same as the sender, userId:{}", userId);
            } else if (isBackendOrder && isUserScoreMaxed(consigneeMemberId, scoreRuleDO)) {
                log.info("Order in shipping listener: The consignee has reached the maximum score, userId:{}, scoreRuleId:{}", consigneeMemberId, scoreRuleDO.getId());
            } else {
                users.add(consigneeMemberId);
            }
        }
        return users;
    }

    /**
     * 校验订单入口是否有效
     */
    private boolean isOrderEntryValid(String orderEntryString, boolean isBackendOrder) {
        if (StringUtils.isBlank(orderEntryString)) {
            log.info("Order in shipping listener: orderEntry is null");
            return false;
        }
        String[] orderEntry = orderEntryString.split(",");
        if (isBackendOrder) {
            return ArrayUtil.contains(orderEntry, String.valueOf(PlatformTypeEnum.BACKEND.getValue()));
        } else {
            return ArrayUtil.contains(orderEntry, String.valueOf(PlatformTypeEnum.APP.getValue())) ||
                    ArrayUtil.contains(orderEntry, String.valueOf(PlatformTypeEnum.WEB.getValue()));
        }
    }

    /**
     * 校验是否为首单且首单是否有效
     */
    private boolean isFirstOrderValid(Long userId, ScoreRuleOrderVExtraVO extraInfo) {
        if (extraInfo.getFirstOrder() == YesOrNoTypeEnum.YES.ordinal()) {
            if (userId == null || userId <= 0) {
                log.info("Order in shipping listener: The user is not member,userId:{}", userId);
                return false;
            }
            if (!orderApi.isFirstOrder(userId)) {
                log.info("Order in shipping listener: Not first order,userId:{}", userId);
                return false;
            }
        }
        return true;
    }

    /**
     * 校验用户积分是否达到上限
     */
    private boolean isUserScoreMaxed(Long userId, ScoreRuleDO scoreRuleDO) {
        if (scoreRuleDO.getMaxScoreTotal() > 0) {
            int userScoreTotalCount = getUserScoreTotalCount(scoreRuleDO.getId(), userId);
            return userScoreTotalCount >= scoreRuleDO.getMaxScoreTotal();
        }
        return false;
    }

    /**
     * 添加符合条件的用户到列表
     */
    private void addValidUser(List<Long> users, Long userId) {
        if (userId != null && userId > 0) {
            users.add(userId);
        }
    }

    /**
     * 根据客户id获取会员id
     *
     * @param customerId
     * @return
     */
    private Long getCustomerMemberId(Long customerId) {
        List<CustomerContactsDO> customerContacts = customerContactsService.getCustomerContactsListByCustomerId(customerId);
        if (!CollectionUtil.isEmpty(customerContacts)) {
            for (CustomerContactsDO customerContact : customerContacts) {
                if (customerContact.getIsDefault() == 1 && customerContact.getUserid() != null) {
                    return customerContact.getUserid();
                }
            }
        }
        log.info("Order in shipping listener: The customer is not bound to the member or have default userId, customerId:{}", customerId);
        return null;
    }

    @Override
    public ScoreRuleTypeEnum getStrategyScoreRuleType() {
        return ScoreRuleTypeEnum.ORDER_V;
    }


    /**
     * 根据规则计算获取积分
     *
     * @param weightOrVolume 重量或体积
     * @param orderVRules    规则
     * @return
     */
    private Integer computeScoreCount(BigDecimal weightOrVolume, List<ScoreRuleOrderVExtraVO.OrderVRule> orderVRules) {
        if (orderVRules.isEmpty()) {
            return 0;
        }
        int wOrV = weightOrVolume.setScale(0, RoundingMode.HALF_UP).intValue();
        ScoreRuleOrderVExtraVO.OrderVRule lastRule = orderVRules.get(orderVRules.size() - 1);
        if (wOrV >= lastRule.getHigh()) {
            return lastRule.getScore();
        }
        for (ScoreRuleOrderVExtraVO.OrderVRule rule : orderVRules) {
            if (rule.getLow() <= wOrV && wOrV < rule.getHigh()) {
                return rule.getScore();
            }
        }
        return 0;
    }
}
